# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.util._data.

The util._data module manages datasets from
https://github.com/chrislit/abydos-data, including downloading them,
decompressing them, and locating them once installed.

Much of this is copied from NLTK's similar facility in
http://www.nltk.org/_modules/nltk/data.html, because they seem to have the
issues figured out, because I don't want to expend the effort to re-invent a
solution, and because their license (Apache) allows for it.
"""

import os
import re
import sys
import urllib.request
import zipfile

from typing import List, Match, Optional, Tuple, cast
from xml.etree import ElementTree  # noqa: S405

__all__ = [
    'data_path',
    'download_package',
    'list_available_packages',
    'list_installed_packages',
    'package_path',
]


DATA_SUBDIRS = ['corpora']
INDEX_URL = (
    'https://raw.githubusercontent.com/chrislit/abydos-data/master/index.xml'
)

data_path = []  # type: List[str]
"""A list of directories where the Abydos data package might reside.
   These directories will be checked in order when looking for a
   resource in the data package.  Note that this allows users to
   substitute in their own versions of resources, if they have them
   (e.g., in their home directory under ~/abydos_data)."""

# User-specified locations:
_paths_from_env = os.environ.get('ABYDOS_DATA', str('')).split(
    os.pathsep
)  # pragma: no cover
data_path += [d for d in _paths_from_env if d]
if 'APPENGINE_RUNTIME' not in os.environ and os.path.expanduser('~/') != '~/':
    data_path.append(os.path.expanduser(str('~/abydos_data')))

if sys.platform.startswith('win'):  # pragma: no cover
    # Common locations on Windows:
    data_path += [
        os.path.join(sys.prefix, str('abydos_data')),
        os.path.join(sys.prefix, str('share'), str('abydos_data')),
        os.path.join(sys.prefix, str('lib'), str('abydos_data')),
        os.path.join(
            os.environ.get(str('APPDATA'), str('C:\\')), str('abydos_data')
        ),
        str(r'C:\abydos_data'),
        str(r'D:\abydos_data'),
        str(r'E:\abydos_data'),
    ]
else:
    # Common locations on UNIX & OS X:
    data_path += [
        os.path.join(sys.prefix, str('abydos_data')),
        os.path.join(sys.prefix, str('share'), str('abydos_data')),
        os.path.join(sys.prefix, str('lib'), str('abydos_data')),
        str('/usr/share/abydos_data'),
        str('/usr/local/share/abydos_data'),
        str('/usr/lib/abydos_data'),
        str('/usr/local/lib/abydos_data'),
    ]


def package_path(resource_name: str) -> str:
    """Given a resource name, returns the path to the package."""
    for path in data_path:
        for subdir in DATA_SUBDIRS:
            check_path = os.path.join(path, subdir, resource_name)
            if os.path.isdir(check_path):
                return check_path
    msg = 'Data package not found. You may need to install or re-install it.'
    raise FileNotFoundError(msg)


def list_installed_packages(
    path: Optional[str] = None,
) -> List[Tuple[str, str, float]]:
    """List all installed data packages."""
    if path:
        paths = [path]
    else:
        paths = data_path
    packages = []
    for path in paths:
        for subdir in DATA_SUBDIRS:
            check_path = os.path.join(path, subdir)
            if os.path.isdir(check_path):
                possible_packages = os.listdir(check_path)
                for package in possible_packages:
                    if os.path.isdir(os.path.join(check_path, package)):
                        with open(
                            os.path.join(check_path, package + '.xml')
                        ) as xml:
                            file = xml.read()
                            name = cast(
                                Match[str], re.search(r'name="([^"]+)"', file)
                            ).group(1)
                            version = cast(
                                Match[str],
                                re.search(r'version="([^"]+)"', file),
                            ).group(1)
                        packages.append((package, name, float(version)))
    return packages


def list_available_packages(
    url: Optional[str] = None,
) -> Tuple[
    List[Tuple[str, str, float, str, str, str]],
    List[Tuple[str, str, List[str]]],
]:
    """List all data packages available for install."""
    installed_packages = {_[0]: _[2] for _ in list_installed_packages()}

    if url is None:
        url = INDEX_URL
    if url[:8] != 'https://':
        raise ValueError('url should begin with "https://"')
    with urllib.request.urlopen(url) as ix:  # noqa: S310
        xml = ElementTree.fromstring(ix.read())  # noqa: S314

    packages = [
        (
            _.attrib['id'],
            _.attrib['name'],
            float(_.attrib['version']),
            _.attrib['url'],
            _.attrib['subdir'],
            'not-installed'
            if _.attrib['id'] not in installed_packages
            else (
                'up-to-date'
                if installed_packages[_.attrib['id']]
                >= float(_.attrib['version'])
                else 'update available'
            ),
        )
        for _ in xml.findall('packages/package')
    ]
    collections = [
        (
            _.attrib['id'],
            _.attrib['name'],
            [__.attrib['ref'] for __ in _.findall('item')],
        )
        for _ in xml.findall('collections/collection')
    ]
    return packages, collections


def _default_download_dir() -> Optional[str]:
    """Return the directory to which packages will be downloaded by default.

    This is mostly copied from NLTK's
    nltk.downloader.Downloader.default_download_dir

    """
    # Check if we are on GAE where we cannot write into filesystem.
    if 'APPENGINE_RUNTIME' in os.environ:  # pragma: no cover
        return None

    # Check if we have sufficient permissions to install in a
    # variety of system-wide locations.
    for abydos_data in data_path:
        if os.path.exists(abydos_data) and os.access(
            abydos_data, os.W_OK
        ):  # pragma: no cover
            return abydos_data

    # On Windows, use %APPDATA%
    if sys.platform == 'win32' and 'APPDATA' in os.environ:  # pragma: no cover
        homedir = os.environ['APPDATA']

    # Otherwise, install in the user's home directory.
    else:  # pragma: no cover
        homedir = os.path.expanduser('~/')
        if homedir == '~/':
            raise ValueError('Could not find a default download directory')

    # append "abydos_data" to the home directory
    return os.path.join(homedir, 'abydos_data')  # pragma: no cover


def download_package(
    resource_name: str,
    url: Optional[str] = None,
    data_path: Optional[str] = None,
    force: bool = False,
    silent: bool = False,
) -> None:
    """Download and install a package or collection."""
    packages, collections = list_available_packages(url)
    installed = list_installed_packages(data_path)
    if data_path is None:
        data_path = _default_download_dir()
    os.makedirs(cast(str, data_path), mode=0o775, exist_ok=True)

    for coll in collections:
        if resource_name == coll[0]:
            if not silent:  # pragma: no branch
                print('Installing {} collection'.format(coll[1]))  # noqa: T001
            for resource_name in coll[2]:
                download_package(resource_name, url, data_path)
            return
    else:
        for pack in packages:
            if resource_name == pack[0]:
                if not force:
                    for inst in installed:  # pragma: no branch
                        if pack[0] == inst[0] and pack[2] <= inst[2]:
                            if not silent:
                                print(  # pragma: no cover  # noqa: T001
                                    '{} package already up-to-date'.format(
                                        pack[1]
                                    )
                                )
                            return
                if not silent:  # pragma: no branch
                    print(  # noqa: T001
                        'Installing {} package'.format(pack[1])
                    )
                zip_fn = os.path.join(
                    cast(str, data_path), pack[4], pack[0] + '.zip'
                )
                os.makedirs(
                    os.path.join(cast(str, data_path), pack[4]),
                    mode=0o775,
                    exist_ok=True,
                )
                urllib.request.urlretrieve(  # noqa: S310
                    pack[3][:-3] + 'xml', zip_fn[:-3] + 'xml'
                )
                urllib.request.urlretrieve(pack[3], zip_fn)  # noqa: S310
                zip_pkg = zipfile.ZipFile(zip_fn)
                zip_pkg.extractall(os.path.join(cast(str, data_path), pack[4]))
                zip_pkg.close()
                os.remove(zip_fn)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._ncd_lzma.

NCD using LZMA
"""

import lzma

from typing import Any

from ._distance import _Distance


__all__ = ['NCDlzma']


class NCDlzma(_Distance):
    """Normalized Compression Distance using LZMA compression.

    Cf. https://en.wikipedia.org/wiki/Lempel-Ziv-Markov_chain_algorithm

    Normalized compression distance (NCD) :cite:`Cilibrasi:2005`.

    .. versionadded:: 0.3.6
    """

    _level = 6

    def __init__(self, level: int = 6, **kwargs: Any) -> None:
        """Initialize LZMA compressor.

        Parameters
        ----------
        level : int
            The compression level (0 to 9)


        .. versionadded:: 0.5.0

        """
        super().__init__(**kwargs)
        self._level = level

    def dist(self, src: str, tar: str) -> float:
        """Return the NCD between two strings using LZMA compression.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            Compression distance

        Examples
        --------
        >>> cmp = NCDlzma()
        >>> cmp.dist('cat', 'hat')
        0.08695652173913043
        >>> cmp.dist('Niall', 'Neil')
        0.16
        >>> cmp.dist('aluminum', 'Catalan')
        0.16
        >>> cmp.dist('ATCG', 'TAGC')
        0.08695652173913043


        .. versionadded:: 0.3.5
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        if src == tar:
            return 0.0

        src_b = src.encode('utf-8')
        tar_b = tar.encode('utf-8')

        src_comp = lzma.compress(src_b, preset=self._level)[14:]
        tar_comp = lzma.compress(tar_b, preset=self._level)[14:]
        concat_comp = lzma.compress(src_b + tar_b, preset=self._level)[14:]
        concat_comp2 = lzma.compress(tar_b + src_b, preset=self._level)[14:]

        return (
            min(len(concat_comp), len(concat_comp2))
            - min(len(src_comp), len(tar_comp))
        ) / max(len(src_comp), len(tar_comp))


if __name__ == '__main__':
    import doctest

    doctest.testmod()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Any

from abydos.distance._distance import _Distance

__all__ = ['AffineGapDistance', 'AffineGapSimilarity']

try:
    from abydos.distance.affinegap import affineGapDistance, normalizedAffineGapDistance
    from abydos.distance.cython_affine import affine, cython_sim_ident
except ImportError:
    import pyximport

    pyximport.install(inplace=True)
    from abydos.distance.affinegap import affineGapDistance, normalizedAffineGapDistance
    from abydos.distance.cython_affine import affine, cython_sim_ident


class AffineGapDistance(_Distance):

    def __init__(self,
                 matchWeight=1,
                 mismatchWeight=11,
                 gapWeight=10,
                 spaceWeight=7,
                 abbreviation_scale=.125):
        super().__init__()
        self.matchWeight = matchWeight
        self.mismatchWeight = mismatchWeight
        self.gapWeight = gapWeight
        self.spaceWeight = spaceWeight
        self.abbreviation_scale = abbreviation_scale

    def dist(self, src: str, tar: str) -> float:
        return affineGapDistance(src,
                                 tar,
                                 self.matchWeight,
                                 self.mismatchWeight,
                                 self.gapWeight,
                                 self.spaceWeight)

    def sim(self, src: str, tar: str) -> float:
        return self.sim_score(src, tar)

    def sim_score(self, src: str, tar: str) -> float:
        return normalizedAffineGapDistance(src,
                                           tar,
                                           self.matchWeight,
                                           self.mismatchWeight,
                                           self.gapWeight,
                                           self.spaceWeight)


class AffineGapSimilarity(_Distance):
    """Returns the affine gap score between two strings.

    The affine gap measure is an extension of the Needleman-Wunsch measure that handles the longer gaps more
    gracefully. For more information refer to the string matching chapter in the DI book ("Principles of Data Integration").

    Args:
        gap_start (float): Cost for the gap at the start (defaults to 1).
        gap_continuation (float): Cost for the gap continuation (defaults to 0.5).
        sim_func (function): Function computing similarity score between two characters, which are represented as strings (defaults
                             to an identity function, which returns 1 if the two characters are the same and returns 0 otherwise).

    Attributes:
        gap_start (float): An attribute to store the gap cost at the start.
        gap_continuation (float): An attribute to store the gap continuation cost.
        sim_func (function): An attribute to store the similarity function.
    """

    def __init__(self, gap_start=1, gap_continuation=0.5, sim_func=cython_sim_ident,
                 **kwargs: Any):
        super().__init__(**kwargs)
        self.gap_start = gap_start
        self.gap_continuation = gap_continuation
        self.sim_func = sim_func

    def dist(self, string1, string2):
        """Computes the affine gap score between two strings. This score can be outside the range [0,1].

        Args:
            string1,string2 (str) : Input strings.

        Returns:
            Affine gap score betwen the two input strings (float).

        Raises:
            TypeError : If the inputs are not strings or if one of the inputs is None.

        # Examples:
        #     >>> aff = Affine()
        #     >>> aff.dist('dva', 'deeva')
        #     1.5
        #     >>> aff = Affine(gap_start=2, gap_continuation=0.5)
        #     >>> aff.dist('dva', 'deeve')
        #     -0.5
        #     >>> aff = Affine(gap_continuation=0.2, sim_func=lambda s1, s2: (int(1 if s1 == s2 else 0)))
        #     >>> aff.dist('AAAGAATTCA', 'AAATCA')
        #     4.4
        # """

        return affine(string1, string2, self.gap_start, self.gap_continuation,
                      self.sim_func)

    def sim_score(self, src: str, tar: str) -> float:
        normalizer = (self.dist(src, src) ** 0.5 * self.dist(tar, tar) ** 0.5)
        if normalizer == 0:
            return 0.0
        return min(max(0.0, self.dist(src, tar)) / normalizer, 1.0)

    def sim(self, src: str, tar: str) -> float:
        return self.sim_score(src, tar)




# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tokenizer._sonoripy.

SonoriPy class
"""

from typing import Callable, Optional, Union

from abydos.tokenizer._tokenizer import _Tokenizer

try:
    from syllabipy.sonoripy import SonoriPy
except ImportError:  # pragma: no cover
    # If the system lacks the SyllabiPy library, that's fine, but SyllabiPy
    # tokenization won't be supported.
    SonoriPy = None  # type: ignore


class SonoriPyTokenizer(_Tokenizer):
    """SonoriPy tokenizer.

    .. versionadded:: 0.4.0
    """

    def __init__(
        self, scaler: Optional[Union[str, Callable[[float], float]]] = None,
    ) -> None:
        """Initialize Tokenizer.

        Parameters
        ----------
        scaler : None, str, or function
            A scaling function for the Counter:

                - None : no scaling
                - 'set' : All non-zero values are set to 1.
                - 'length' : Each token has weight equal to its length.
                - 'length-log' : Each token has weight equal to the log of its
                   length + 1.
                - 'length-exp' : Each token has weight equal to e raised to its
                   length.
                - a callable function : The function is applied to each value
                  in the Counter. Some useful functions include math.exp,
                  math.log1p, math.sqrt, and indexes into interesting integer
                  sequences such as the Fibonacci sequence.


        .. versionadded:: 0.4.0

        """
        if SonoriPy is None:
            raise TypeError(  # pragma: no cover
                'SonoriPy tokenizer requires installation of SyllabiPy'
                + ' package.'
            )

        super(SonoriPyTokenizer, self).__init__(scaler)

    def tokenize(self, string: str) -> 'SonoriPyTokenizer':
        """Tokenize the term and store it.

        The tokenized term is stored as an ordered list and as a Counter
        object.

        Parameters
        ----------
        string : str
            The string to tokenize

        Examples
        --------
        >>> SonoriPyTokenizer().tokenize('seven-twelfths')
        SonoriPyTokenizer({'se': 1, 'ven-': 1, 'twelfths': 1})

        >>> SonoriPyTokenizer().tokenize('character')
        SonoriPyTokenizer({'cha': 1, 'rac': 1, 'ter': 1})


        .. versionadded:: 0.4.0

        """
        self._string = string

        self._ordered_tokens = []
        for word in string.split():
            self._ordered_tokens += SonoriPy(word)
        if not self._ordered_tokens:
            self._ordered_tokens = [self._string]

        self._scale_and_counterize()
        return self


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""setup.py.

setuptools configuration file for Abydos
"""
import importlib
import os
from codecs import open
from os import path
from pathlib import Path

from setuptools import find_packages, setup

HERE = Path(path.abspath(path.dirname(__file__)))

__package_name__ = 'abydos'
__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


def import_file(module_name, file_path):
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def get_version(package_name):
    version = import_file('version',
                          os.path.join(__location__, package_name, 'version.py'))
    return version.__version__


__version__ = get_version(__package_name__)


def readfile(fn):
    return (HERE / fn).read_text(encoding='utf8')


def read_requirements(reqs_path):
    with open(reqs_path, encoding='utf8') as f:
        reqs = [line.strip() for line in f
                if not line.strip().startswith('#') and not line.strip().startswith('--')]
    return reqs


def get_extra_requires(path, add_all=True, ext='*.txt'):
    main_reqs = read_requirements(HERE / 'requirements.txt')

    extra_deps = {}
    for filename in path.glob(ext):
        # convention of naming requirements files: requirements.{module}.txt
        package_suffix = filename.suffixes[-2].strip('.')
        reqs = list({*main_reqs, *read_requirements(filename)})
        extra_deps[package_suffix] = reqs

    if add_all:
        extra_deps['all'] = set(vv for v in extra_deps.values() for vv in v)
    return extra_deps

try:
    from setuptools import Extension
    from Cython.Build import cythonize
    use_cython = True
except ImportError:
    use_cython = False

if use_cython:
    ext_modules = cythonize([Extension('abydos.distance.affinegap', ['abydos/distance/affinegap.pyx']),
                             Extension('abydos.distance.cython_affine', ['abydos/distance/cython_affine.pyx'])])
else:
    ext_modules = [Extension('abydos.distance.affinegap', ['abydos/distance/affinegap.c']),
                   Extension('abydos.distance.cython_affine', ['abydos/distance/cython_affine.c'])]

setup(
    name='aj_abydos_mod',
    packages=find_packages(exclude=['tests*']),
    version=__version__,
    description='Fork of the Abydos NLP/IR library',
    author='Christopher C. Little',
    author_email='chrisclittle+abydos@gmail.com',
    ext_modules=ext_modules,
    url='https://github.com/anuj1508/abydos',
    download_url='https://github.com/anuj1508/abydos/archive/master.zip',
    keywords=[
        'nlp',
        'ai',
        'ir',
        'language',
        'linguistics',
        'phonetic algorithms',
        'string distance',
    ],
    license='GPLv3+',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or \
later (GPLv3+)',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Indexing',
        'Topic :: Text Processing :: Linguistic',
        'Natural Language :: English',
    ],
    long_description_content_type='text/x-rst',
    long_description='\n\n'.join(
        [readfile(f) for f in ('README.rst', 'HISTORY.rst', 'AUTHORS.rst')]
    ),
    python_requires='>=3.8',
    install_requires=read_requirements(HERE / 'requirements.txt'),
    extras_require=get_extra_requires(HERE / 'requirements')
)



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._batagelj_bren.

Batagelj & Bren distance
"""

from typing import Any, Counter as TCounter, Optional, Sequence, Set, Union

from abydos.distance._token_distance import _TokenDistance
from abydos.tokenizer import _Tokenizer

__all__ = ['BatageljBren']


class BatageljBren(_TokenDistance):
    r"""Batagelj & Bren distance.

    For two sets X and Y and a population N, the Batagelj & Bren
    distance :cite:`Batagelj:1995`, Batagelj & Bren's :math:`Q_0`, is

        .. math::

            dist_{BatageljBren}(X, Y) =
            \frac{|X \setminus Y| \cdot |Y \setminus X|}
            {|X \cap Y| \cdot |(N \setminus X) \setminus Y|}

    In :ref:`2x2 confusion table terms <confusion_table>`, where a+b+c+d=n,
    this is

        .. math::

            dist_{BatageljBren} =
            \frac{bc}{ad}

    .. versionadded:: 0.4.0
    """

    def __init__(
        self,
        alphabet: Optional[
            Union[TCounter[str], Sequence[str], Set[str], int]
        ] = None,
        tokenizer: Optional[_Tokenizer] = None,
        intersection_type: str = 'crisp',
        **kwargs: Any
    ) -> None:
        """Initialize BatageljBren instance.

        Parameters
        ----------
        alphabet : Counter, collection, int, or None
            This represents the alphabet of possible tokens.
            See :ref:`alphabet <alphabet>` description in
            :py:class:`_TokenDistance` for details.
        tokenizer : _Tokenizer
            A tokenizer instance from the :py:mod:`abydos.tokenizer` package
        intersection_type : str
            Specifies the intersection type, and set type as a result:
            See :ref:`intersection_type <intersection_type>` description in
            :py:class:`_TokenDistance` for details.
        **kwargs
            Arbitrary keyword arguments

        Other Parameters
        ----------------
        qval : int
            The length of each q-gram. Using this parameter and tokenizer=None
            will cause the instance to use the QGram tokenizer with this
            q value.
        metric : _Distance
            A string distance measure class for use in the ``soft`` and
            ``fuzzy`` variants.
        threshold : float
            A threshold value, similarities above which are counted as
            members of the intersection for the ``fuzzy`` variant.


        .. versionadded:: 0.4.0

        """
        super(BatageljBren, self).__init__(
            alphabet=alphabet,
            tokenizer=tokenizer,
            intersection_type=intersection_type,
            **kwargs
        )

    def dist_abs(self, src: str, tar: str) -> float:
        """Return the Batagelj & Bren distance of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Batagelj & Bren distance

        Examples
        --------
        >>> cmp = BatageljBren()
        >>> cmp.dist_abs('cat', 'hat')
        0.002570694087403599
        >>> cmp.dist_abs('Niall', 'Neil')
        0.007741935483870968
        >>> cmp.dist_abs('aluminum', 'Catalan')
        0.07282184655396619
        >>> cmp.dist_abs('ATCG', 'TAGC')
        inf


        .. versionadded:: 0.4.0

        """
        if src == tar:
            return 0.0

        self._tokenize(src, tar)

        a = self._intersection_card()
        b = self._src_only_card()
        c = self._tar_only_card()
        d = self._total_complement_card()

        if a == 0 or d == 0:
            return float('inf')
        return b * c / (a * d)

    def dist(self, src: str, tar: str) -> float:
        """Return the normalized Batagelj & Bren distance of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Normalized Batagelj & Bren distance

        Examples
        --------
        >>> cmp = BatageljBren()
        >>> cmp.dist('cat', 'hat')
        3.2789465400556106e-06
        >>> cmp.dist('Niall', 'Neil')
        9.874917709019092e-06
        >>> cmp.dist('aluminum', 'Catalan')
        9.276668350823718e-05
        >>> cmp.dist('ATCG', 'TAGC')
        1.0


        .. versionadded:: 0.4.0

        """
        if src == tar:
            return 0.0

        self._tokenize(src, tar)

        a = self._intersection_card()
        b = self._src_only_card()
        c = self._tar_only_card()
        d = self._total_complement_card()

        if a == 0 or d == 0:
            return 1.0
        return (b * c / (a * d)) / (a + b + c + d)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._gwet_ac.

Gwet's AC correlation
"""

from typing import Any, Counter as TCounter, Optional, Sequence, Set, Union

from abydos.distance._token_distance import _TokenDistance
from abydos.tokenizer import _Tokenizer

__all__ = ['GwetAC']


class GwetAC(_TokenDistance):
    r"""Gwet's AC correlation.

    For two sets X and Y and a population N, Gwet's AC correlation
    :cite:`Gwet:2008` is

        .. math::

            corr_{Gwet_{AC}}(X, Y) = AC =
            \frac{p_o - p_e^{AC}}{1 - p_e^{AC}}

    where

        .. math::

            \begin{array}{lll}
            p_o &=&\frac{|X \cap Y| + |(N \setminus X) \setminus Y|}{|N|}

            p_e^{AC}&=&\frac{1}{2}\Big(\frac{|X|+|Y|}{|N|}\cdot
            \frac{|X \setminus Y| + |Y \setminus X|}{|N|}\Big)
            \end{array}


    In :ref:`2x2 confusion table terms <confusion_table>`, where a+b+c+d=n,
    this is

        .. math::

            \begin{array}{lll}
            p_o&=&\frac{a+d}{n}

            p_e^{AC}&=&\frac{1}{2}\Big(\frac{2a+b+c}{n}\cdot
            \frac{2d+b+c}{n}\Big)
            \end{array}

    .. versionadded:: 0.4.0
    """

    def __init__(
        self,
        alphabet: Optional[
            Union[TCounter[str], Sequence[str], Set[str], int]
        ] = None,
        tokenizer: Optional[_Tokenizer] = None,
        intersection_type: str = 'crisp',
        **kwargs: Any
    ) -> None:
        """Initialize GwetAC instance.

        Parameters
        ----------
        alphabet : Counter, collection, int, or None
            This represents the alphabet of possible tokens.
            See :ref:`alphabet <alphabet>` description in
            :py:class:`_TokenDistance` for details.
        tokenizer : _Tokenizer
            A tokenizer instance from the :py:mod:`abydos.tokenizer` package
        intersection_type : str
            Specifies the intersection type, and set type as a result:
            See :ref:`intersection_type <intersection_type>` description in
            :py:class:`_TokenDistance` for details.
        **kwargs
            Arbitrary keyword arguments

        Other Parameters
        ----------------
        qval : int
            The length of each q-gram. Using this parameter and tokenizer=None
            will cause the instance to use the QGram tokenizer with this
            q value.
        metric : _Distance
            A string distance measure class for use in the ``soft`` and
            ``fuzzy`` variants.
        threshold : float
            A threshold value, similarities above which are counted as
            members of the intersection for the ``fuzzy`` variant.


        .. versionadded:: 0.4.0

        """
        super(GwetAC, self).__init__(
            alphabet=alphabet,
            tokenizer=tokenizer,
            intersection_type=intersection_type,
            **kwargs
        )

    def corr(self, src: str, tar: str) -> float:
        """Return the Gwet's AC correlation of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Gwet's AC correlation

        Examples
        --------
        >>> cmp = GwetAC()
        >>> cmp.corr('cat', 'hat')
        0.9948456319360438
        >>> cmp.corr('Niall', 'Neil')
        0.990945276504824
        >>> cmp.corr('aluminum', 'Catalan')
        0.9804734301840141
        >>> cmp.corr('ATCG', 'TAGC')
        0.9870811678360627


        .. versionadded:: 0.4.0

        """
        if src == tar:
            return 1.0

        self._tokenize(src, tar)

        a = self._intersection_card()
        b = self._src_only_card()
        c = self._tar_only_card()
        d = self._total_complement_card()
        n = a + b + c + d

        po = (a + d) / n
        q = (2 * a + b + c) / (2 * n)
        pe = 2 * q * (1 - q)

        return (po - pe) / (1 - pe)

    def sim(self, src: str, tar: str) -> float:
        """Return the Gwet's AC similarity of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Gwet's AC similarity

        Examples
        --------
        >>> cmp = GwetAC()
        >>> cmp.sim('cat', 'hat')
        0.9974228159680218
        >>> cmp.sim('Niall', 'Neil')
        0.995472638252412
        >>> cmp.sim('aluminum', 'Catalan')
        0.9902367150920071
        >>> cmp.sim('ATCG', 'TAGC')
        0.9935405839180314


        .. versionadded:: 0.4.0

        """
        return (1.0 + self.corr(src, tar)) / 2.0


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.fingerprint._position.

Cisłak & Grabowski's position fingerprint
"""

from typing import Dict, Tuple

from abydos.fingerprint._fingerprint import MOST_COMMON_LETTERS_CG, _Fingerprint

__all__ = ['Position']


class Position(_Fingerprint):
    """Position Fingerprint.

    Based on the position fingerprint from :cite:`Cislak:2017`.

    .. versionadded:: 0.3.6
    """

    def __init__(
        self,
        n_bits: int = 16,
        most_common: Tuple[str, ...] = MOST_COMMON_LETTERS_CG,
        bits_per_letter: int = 3,
    ) -> None:
        """Initialize Count instance.

        Parameters
        ----------
        n_bits : int
            Number of bits in the fingerprint returned
        most_common : list
            The most common tokens in the target language, ordered by frequency


        .. versionadded:: 0.4.0

        """
        super(Position, self).__init__()
        self._n_bits = n_bits
        self._most_common = most_common
        self._bits_per_letter = bits_per_letter

    def fingerprint(self, word: str) -> str:
        """Return the position fingerprint.

        Parameters
        ----------
        word : str
            The word to fingerprint

        Returns
        -------
        str
            The position fingerprint

        Examples
        --------
        >>> pf = Position()
        >>> pf.fingerprint('hat')
        '1110100011111111'
        >>> pf.fingerprint('niall')
        '1111110101110010'
        >>> pf.fingerprint('colin')
        '1111111110010111'
        >>> pf.fingerprint('atcg')
        '1110010001111111'
        >>> pf.fingerprint('entreatment')
        '0000101011111111'


        .. versionadded:: 0.3.0
        .. versionchanged:: 0.3.6
            Encapsulated in class
        .. versionchanged:: 0.6.0
            Changed to return a str and added fingerprint_int method

        """
        return ('{:0' + str(self._n_bits) + 'b}').format(
            self.fingerprint_int(word)
        )

    def fingerprint_int(self, word: str) -> int:
        """Return the position fingerprint.

        Parameters
        ----------
        word : str
            The word to fingerprint

        Returns
        -------
        int
            The position fingerprint as an int

        Examples
        --------
        >>> pf = Position()
        >>> pf.fingerprint_int('hat')
        59647
        >>> pf.fingerprint_int('niall')
        64882
        >>> pf.fingerprint_int('colin')
        65431
        >>> pf.fingerprint_int('atcg')
        58495
        >>> pf.fingerprint_int('entreatment')
        2815


        .. versionadded:: 0.6.0

        """
        n_bits = self._n_bits
        position = {}  # type: Dict[str, int]
        for pos, letter in enumerate(word):
            if letter not in position and letter in self._most_common:
                position[letter] = min(pos, 2 ** self._bits_per_letter - 1)

        fingerprint = 0

        for letter in self._most_common:
            if n_bits:
                fingerprint <<= min(self._bits_per_letter, n_bits)
                if letter in position:
                    fingerprint += min(position[letter], 2 ** n_bits - 1)
                else:
                    fingerprint += min(
                        2 ** self._bits_per_letter - 1, 2 ** n_bits - 1
                    )
                n_bits -= min(self._bits_per_letter, n_bits)
            else:
                break

        for _ in range(n_bits):
            fingerprint <<= 1
            fingerprint += 1

        return fingerprint


if __name__ == '__main__':
    import doctest

    doctest.testmod()
